import json
import os
from shapely.geometry import MultiPoint, LineString, mapping
from ibm2ieee import ibm2float32
import numpy as np


class segy(object):
    file_path = os.path.dirname(__file__)
    structure_file = "assets/segyr1_structure.json"
    config_path = os.path.join(file_path, structure_file)
    custom_config = ""  # file_path of the overly config
    pass

    def __init__(self, flo, **kwargs):
        self.__dict__.update(kwargs)
        self.flo = flo  # File Like Object
        self.get_config()
        self.__get_header()
        self.__get_binary_header()
        self.__get_aux_headers()
        self.__set_data_type()

    def get_config(self):
        self.config = json.load(open(self.config_path, "r"))
        if self.custom_config != "":
            custom = json.load(open(self.custom_config, "r"))
            self.config = self.__update(self.config, custom)
        self.byte_order = self.config['byte_order']
        self.byte_order_symbol = ">" if self.byte_order == 'big' else ">"

    def __update(self, d, u):
        for k, v in u.items():
            if isinstance(v, dict):
                d[k] = self.__update(d.get(k, {}), v)
            else:
                d[k] = v
        return d

    def __get_header(self, headers=0):
        if headers == 0:
            # normal headers from rev 0
            self.header = []
            bytes_to_read = 3200
        else:
            # extened headers optional rev1
            self.flo.seek(3600, 0)
            bytes_to_read = 3200 * headers
        bytes = self.flo.read(bytes_to_read)
        textEBCDIC = bytes.decode('cp500')
        textascii = bytes.decode('latin1')
        if textEBCDIC.count(" ") >= textascii.count(' '):
            correctHeader = textEBCDIC
        else:
            correctHeader = textascii
        for i in range(int(bytes_to_read / 80)):  # read each line of 80
            self.header.append(correctHeader[i * 80:(i + 1) * 80].strip())

    def __get_binary_header(self):
        self.flo.seek(0, 0)
        # self.flo.seek(3200,0)
        self.flo.read(3200)
        bin_config = self.config['binary_header']
        bytes = self.flo.read(bin_config['bytes'])
        self.bin_header = {}
        for field, value in bin_config['fields'].items():
            value = self.__get_value(value, bytes)
            self.bin_header[field] = value
        self.trace_samples = (self.bin_header['samples_per_trace'])
        self.trace_sample_interval = int(self.bin_header['sample_interval'] / 1000)
        self.trace_z_range = range(0, int(self.trace_samples * self.trace_sample_interval), self.trace_sample_interval)

    def __get_aux_headers(self):
        self.header_bytes = 3600
        if self.bin_header['extended_header_count'] > 0:
            self.__get_header(self.bin_header['extended_header_count'])
            self.header_bytes += int(self.bin_header['extended_header_count'] * 3200)
        # Todo process the data sections that are documented

    def __set_data_type(self):
        code = self.bin_header['data_sample_format_code']
        if code == 1:
            self.data_bytes = 4
            self.data_format = "ibm"
            self.extract_trace_data = self.__ibm4_decoder
        elif code == 2:
            self.data_bytes = 4
            self.data_format = "int4"
            self.extract_trace_data = self.__int4_decoder
        elif code == 3:
            self.data_bytes = 2
            self.data_format = "int2"
            self.extract_trace_data = self.__int2_decoder
        elif code == 5:
            self.data_bytes = 4
            self.data_format = "ieee",
            self.extract_trace_data = self.__ieee4_decoder
        elif code == 8:
            self.data_bytes = 1
            self.data_format = "int1",
            self.extract_trace_data = self.__int1_decoder
        else:
            self.data_bytes = 4
            self.data_format = "int4",
            self.extract_trace_data = self.__int4_decoder
        self.trace_bytes = self.bin_header['samples_per_trace'] * self.data_bytes
        self.trace_count = (self.flo.seek(0, 2) - self.header_bytes - 1500) / (240 + self.trace_bytes)
        trace = self.get_trace(0)
        scalar_raw = trace[0]['scalar_coordinate']
        if scalar_raw == 0:
            self.scalar = 1
        elif scalar_raw > 0:
            self.scalar = scalar_raw
        else:
            self.scalar = 1 / abs(scalar_raw)

    def get_trace(self, index, sample_mask=[]):
        if index > self.trace_count or index < -self.trace_count:
            return "error", []
        elif (index < 0):
            index = self.trace_count + index + 1
        # trace index starts at zero
        seek_bytes = int((self.trace_bytes + 240) * index)
        self.flo.seek(self.header_bytes + seek_bytes, 0)
        header_raw = self.flo.read(240)
        trace_raw = self.flo.read(self.trace_bytes)
        traceheader = self.__extract_trace_header(header_raw)
        tracedata = self.extract_trace_data(trace_raw)
        if len(sample_mask) != 0:
            tracedata = tracedata[sample_mask]
        return traceheader, tracedata

    def get_traces(self, **kwargs):
        options = {'inline': [-np.inf, np.inf],
                   'crossline': [-np.inf, np.inf],
                   'z_range': [-np.inf, np.inf],
                   'reverse_order': False}
        options.update(kwargs)

        # Create bitmask filters
        filteril = (self.inline >= np.min(options['inline'])) & (self.inline <= np.max(options['inline']))
        filterxl = (self.crossline >= np.min(options['crossline'])) & (self.crossline <= np.max(options['crossline']))
        filterz = ((self.trace_z_range >= np.min(options['z_range'])) & (
                    self.trace_z_range <= np.max(options['z_range'])))
        trace_ids = self.trace_indexes[filteril & filterxl]
        filterskips = np.array(range(len(trace_ids))) % options['skip'] == 0
        trace_ids = trace_ids[filterskips]

        if (options['reverse_order']):
            trace_ids = trace_ids[::-1]

        # Get only data that is needed
        outtraces = []
        outheaders = []
        for id in trace_ids:
            result = self.get_trace(id, filterz)
            outtraces.append(result[1])
            outheaders.append(result[0])

        # Return Results
        z_range = np.array(self.trace_z_range)[filterz]
        z_range = range(np.min(z_range), np.max(z_range), z_range[1] - z_range[0])
        return outheaders, np.array(outtraces), z_range

    def __extract_trace_header(self, header_raw):
        trace_config = self.config['trace_header']
        trace_header = {}
        for field, value in trace_config['fields'].items():
            value = self.__get_value(value, header_raw)
            trace_header[field] = value
        return trace_header

    def scan_trace_headers(self):
        self.flo.seek(self.header_bytes, 0)
        self.trace_indexes = np.array(range(int(self.trace_count + 1)))
        inline = []
        crossline = []
        x_cord = []
        y_cord = []
        for i in self.trace_indexes:
            header_raw = self.flo.read(240)
            self.flo.seek(self.trace_bytes, 1)
            header = self.__extract_trace_header(header_raw)
            inline.append(header['inline'])
            crossline.append(header['crossline'])
            x_cord.append(header['cdp_x'] * self.scalar)
            y_cord.append(header['cdp_y'] * self.scalar)
        self.inline = np.array(inline)
        self.crossline = np.array(crossline)
        self.x_cord = np.array(x_cord)
        self.y_cord = np.array(y_cord)

    def get_3d_parameters(self):
        # create a line / crossline  matrix for calculating geometry
        inline_crossline = np.array([self.inline, self.crossline, np.ones(len(self.inline))]).transpose()
        xypoints = np.array([self.x_cord, self.y_cord, np.ones(len(self.inline))]).transpose()
        # calculate the affine transform matrix
        try:
            z = np.linalg.lstsq(inline_crossline, xypoints, rcond=None)
            transform = z[0]
        except:
            transform = False
        # use the points to make a polygon around the survey and save as well known text
        outline = mapping(MultiPoint(xypoints).convex_hull)
        # compile the return dictonary file
        self.dimensions = 3
        self.transform = transform
        self.geom = outline

        inline_mesh, crossline_mesh = np.meshgrid(range(self.inline.min(), self.inline.max()),
                                                  range(self.crossline.min(), self.crossline.max()))

        self.inline_mesh = inline_mesh.transpose()
        self.crossline_mesh = crossline_mesh.transpose()

    def get_2d_parameters(self):
        # get line trace ranges
        xypoints = np.array([self.x_cord, self.y_cord]).transpose()
        geom = mapping(LineString(xypoints))
        self.dimensions = 2
        self.geom = geom

    def close(self):
        self.flo.close()

    def __int1_decoder(self, raw):
        format = self.byte_order_symbol + "i1"
        data = np.frombuffer(raw, dtype=format)
        return data

    def __int2_decoder(self, raw):
        format = self.byte_order_symbol + "i2"
        data = np.frombuffer(raw, dtype=format)
        return data

    def __int4_decoder(self, raw):
        format = self.byte_order_symbol + "i4"
        data = np.frombuffer(raw, dtype=format)
        return data

    def __ieee4_decoder(self, raw):
        format = self.byte_order_symbol + "f4"
        data = np.frombuffer(raw, dtype=format)
        return data

    def __ibm4_decoder(self, raw):
        format = self.byte_order_symbol + "u4"
        data = np.frombuffer(raw, dtype=format)
        data = ibm2float32(data)
        return data

    def __get_value(self, field, data):
        if "constant" in field:
            return field['constant']
        try:
            pos = field['start_byte'] - 1
            len = field['length_bytes']
            return int.from_bytes(data[pos:pos + len], byteorder='big', signed=True)
        except:
            return 0

    def xy2il(self, xs, ys):
        matrix = np.array([xs, ys, np.ones(len(xs))]).transpose()
        tint = np.linalg.inv(self.transform)
        ic = np.matmul(matrix, tint).astype(int)
        return ic[:, 0:2]

    def ic2xy(self, inlines, crosslines):
        matrix = np.array([inlines, crosslines, np.ones(len(inlines))]).transpose()
        xy = np.matmul(matrix, self.transform)
        return xy[:, 0:2]



import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import io
import base64
import copy


def plotSeismic(segy_file, **kwargs):
    options = {
        'inline': [-np.inf, np.inf],
        'crossline': [-np.inf, np.inf],
        'z_range': [-np.inf, np.inf],
        'tmpdir': 'tmp/',
        'dpi': 160,
        'reverse_order': False,
        "color_map": "seismic",
        "variable_density": False,
        "wiggle": True,
        "variable_area": True,
        "skip": 1,
        "base64out": False,
        "fileout": False,
        "filelikeobjectout": True}

    options.update(kwargs)
    # key words include
    # in_line = [int, int]
    # x_line = [int, int]
    # z_range = [int, int] time(ms) or depth(units)
    # reverse_order = True/False
    # variable_density = True/False
    # color_map = matplotLlib color map text https://matplotlib.org/gallery/color/colormap_reference.html

    segy_file.scan_trace_headers()
    headers, traces, z_range = segy_file.get_traces(**options)

    fig = plotSeismicLine(traces, **options)
    section_file = io.BytesIO()
    fig.savefig(section_file, format='png', dpi=options['dpi'], antialiased=1)
    del fig

    fig = plotFrequencies(traces, segy_file.bin_header['sample_interval'], 5)
    freq_file = io.BytesIO()
    fig.savefig(freq_file, format='png', dpi=options['dpi'], antialiased=1)
    del fig

    result = {}
    if options['filelikeobjectout']:
        section_file.seek(0)
        freq_file.seek(0)
        result['section_flo'] = section_file
        result['freq_flo'] = freq_file
    if options['fileout']:
        section_file.seek(0)
        freq_file.seek(0)
        open(options['tmpdir'] + "section.png", 'wb').write(section_file.read())
        result['section_file'] = options['tmpdir'] + "section.png"
        open(options['tmpdir'] + "frequency.png", 'wb').write(freq_file.read())
        result['freq_file'] = options['tmpdir'] + "frequency.png"
    if options['base64out']:
        section_file.seek(0)
        freq_file.seek(0)
        result['section_64'] = base64.b64encode(section_file.read()).decode()
        result['freq_64'] = base64.b64encode(freq_file.read()).decode()

    return result


def plotSeismicLine(values_matrix, **kwargs):
    options = {'x_range': range(values_matrix.shape[0]),
               'y_range': range(values_matrix.shape[1]),
               'color_map': "seismic",
               'line_width': .5,
               'gain': 3,
               'color': 'black'
               }
    options.update(kwargs)

    # y, x = np.mgrid[options['y_range'],options['x_range']]

    z = np.transpose(values_matrix)
    zmax = np.percentile(z, 98)
    zmin = np.percentile(z, 2)
    if zmin == zmax:
        zmin = -1
        zmax = 1
    norm = colors.Normalize(vmin=zmin, vmax=zmax)
    # fig, (ax0, ax1) = plt.subplots(nrows=2)
    fig = plt.figure(figsize=[16, 9])
    ax0 = fig.add_axes([0, 0, 1, 1])
    if (options['variable_density']):
        ax0.imshow(z, origin='upper', norm=norm, cmap=options['color_map'], interpolation='bicubic', aspect="auto")
        return fig
    else:
        gain = options['gain']
        line_width = options['line_width']
        ns = values_matrix.shape[1]
        ntraces = values_matrix.shape[0]
        max_traces = 1500
        step = 1
        if max_traces < ntraces:
            step = int(np.ceil(ntraces / max_traces))

        samples = np.array(range(ns))

        for i in range(0, ntraces, step):
            # use copy to avoid truncating the data
            trace = copy.copy(values_matrix[i, :])
            trace[0] = 0
            trace[-1] = 0
            trace = trace * gain / zmax
            ax0.plot(i + trace, samples, color=options['color'], linewidth=line_width)
            if (options["variable_area"]):
                filltrace = copy.copy(trace)
                filltrace[trace < 0] = 0
                ax0.fill(i + filltrace, samples, 'k', linewidth=0, color=options['color'])

    ax0.grid(True)
    ax0.invert_yaxis()
    plt.ylim([np.max(samples), np.min(samples)])
    plt.xlim([ntraces, 0])
    return plt


def plotFrequencies(z_matrix, sample_interval, windows=1):
    zlen = z_matrix.shape[1]
    window_len = zlen / windows
    ubound = 1000 / sample_interval / 2  # nyquest assuming no interpolation

    fig = plt.figure(figsize=[12, 9])
    ax0 = fig.add_axes([.1, .1, .8, .8])
    freq = []
    amps_raw = []
    amps_db = []
    for i in range(windows):
        z_subset = z_matrix[:, int(i * window_len):int((i + 1) * window_len)]
        window_freq, window_amps_raw, window_amps_db = getfrequencies(z_subset, sample_interval)
        freq.append(window_freq)
        amps_raw.append(window_amps_raw),
        amps_db.append(window_amps_db)
        label = ("Window " + str(i + 1))
        ax0.plot(freq[i], amps_db[i], label=("Window " + str(i + 1)))

    ax0.set_xlim(0, ubound)
    ax0.set_ylim(-50, 0)
    ax0.legend()
    plt.xlabel('Hz or Cycles/1000 Survey Units')
    plt.ylabel('DB Down')

    return fig


def getfrequencies(z_matrix, sample_rate):
    # assumes that the first cordinate is a spatial cordiantate and the second is time.
    # run the process on all time cordinates then stackes them to get good average numbers.
    raw = np.fft.fft(z_matrix, axis=1)
    amps_raw = np.abs(raw)
    av_amps = amps_raw.mean(axis=0)
    max_amp = np.max(av_amps)
    freq = np.fft.fftfreq(len(av_amps), sample_rate / 1000)
    amps_DB_down = 20 * np.log10(av_amps / max_amp)
    filter = (freq >= 0)
    freq = freq[filter]
    av_amps = av_amps[filter]
    amps_DB_down = amps_DB_down[filter]
    return freq, av_amps, amps_DB_down


def interpolate(z_matrix, current_sample_rate, requested_sample_rate):
    # This resamples the data along the second dimension of the matrix
    # Check if necessary
    if current_sample_rate == requested_sample_rate:
        return z_matrix
    # If needed inport library and run resample
    import scipy.signal as sig
    new_samples = int(z_matrix.shape[1] * current_sample_rate / requested_sample_rate)
    out_matrix = sig.resample(z_matrix, new_samples, axis=1)
    return out_matrix



import setuptools

long_description = open("README.md", "r").read()

setuptools.setup(
    name="segy_lite",
    version="0.0.4",
    author="Aaron Close",
    author_email="adclose@gmail.com",
    description="Lightweight SEGY Reading and Plotting Tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=['numpy', 'scipy', 'matplotlib', 'ibm2ieee', 'ebcdic', 'shapely', 'descartes'],
    url="https://github.com/adclose/segy_lite",
    packages=setuptools.find_packages(exclude='tests'),
    package_data={"": ["assets/*.json"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: GIS",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)